@@ -51,10 +51,11 @@ pub struct CertStore<'a> {
     certd: std::result::Result<store::CertD<'a>, store::Certs<'a>>,
 
     // Read-only backends.
-    backends: Vec<(Box<dyn store::Store<'a> + 'a>, AccessMode)>,
+    backends: Vec<(Box<dyn store::Store<'a> + Send + Sync + 'a>, AccessMode)>,
 
     keyserver: Option<Box<store::KeyServer<'a>>>,
 }
+assert_send_and_sync!(CertStore<'_>);
 
 impl<'a> CertStore<'a> {
     /// Returns a CertStore, which does not have any configured backends.
@@ -127,7 +128,7 @@ impl<'a> CertStore<'a> {
     /// Add the specified backend to the CertStore.
     ///
     /// The backend is added to the collection of read-only backends.
-    pub fn add_backend(&mut self, backend: Box<dyn store::Store<'a> + 'a>,
+    pub fn add_backend(&mut self, backend: Box<dyn store::Store<'a> + Send + Sync + 'a>,
                        mode: AccessMode)
         -> &mut Self
     {
@@ -178,7 +179,7 @@ impl<'a> CertStore<'a> {
     where P: AsRef<Path>,
           I: IntoIterator<Item=P>,
     {
-        let mut keyring = Certs::empty();
+        let keyring = Certs::empty();
         let mut error = None;
         for filename in filenames {
             let filename = filename.as_ref();
@@ -558,36 +559,36 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         Box::new(certs.into_iter())
     }
 
-    fn prefetch_all(&mut self) {
-        match self.certd.as_mut() {
+    fn prefetch_all(&self) {
+        match self.certd.as_ref() {
             Ok(certd) => certd.prefetch_all(),
             Err(in_memory) => in_memory.prefetch_all(),
         };
 
-        for (backend, _mode) in self.backends.iter_mut() {
+        for (backend, _mode) in self.backends.iter() {
             backend.prefetch_all();
         }
     }
 
-    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
-        match self.certd.as_mut() {
+    fn prefetch_some(&self, certs: &[KeyHandle]) {
+        match self.certd.as_ref() {
             Ok(certd) => certd.prefetch_some(certs),
             Err(in_memory) => in_memory.prefetch_some(certs),
         };
 
-        for (backend, _mode) in self.backends.iter_mut() {
+        for (backend, _mode) in self.backends.iter() {
             backend.prefetch_some(certs);
         }
     }
 }
 
 impl<'a> store::StoreUpdate<'a> for CertStore<'a> {
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                 merge_strategy: &mut dyn MergeCerts<'a>)
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>
     {
         tracer!(TRACE, "CertStore::update_by");
-        match self.certd.as_mut() {
+        match self.certd.as_ref() {
             Ok(certd) => {
                 t!("Forwarding to underlying certd");
                 certd.update_by(cert, merge_strategy)
@@ -654,7 +655,7 @@ impl<'a> CertStore<'a> {
         // Now purge the successfully flushed certs from the keyserver
         // cache, so that we will not try to flush them again next
         // time we're called.
-        flushed_certs.iter().for_each(|c| ks.delete_from_cache(c));
+        ks.delete_from_cache(flushed_certs.iter().map(Arc::clone));
 
         t!("Flushed {} certificates", flushed_certs.len());
         result
