@@ -31,6 +31,7 @@ pub struct LazyCert<'a> {
     raw: OnceLock<RawCert<'a>>,
     cert: OnceLock<Cow<'a, Cert>>,
 }
+assert_send_and_sync!(LazyCert<'_>);
 
 impl<'a> std::fmt::Debug for LazyCert<'a> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
@@ -182,8 +183,10 @@ impl<'a> LazyCert<'a> {
             t!("Resolving {}", raw.fingerprint());
             match Cert::try_from(raw) {
                 Ok(cert) => {
-                    self.cert.set(Cow::Owned(cert))
-                        .expect("just checked that it was empty");
+                    // The following will fail, if we lost a race to
+                    // parse the raw certificate.  That's okay,
+                    // because parsing is deterministic.
+                    let _ = self.cert.set(Cow::Owned(cert));
                 }
                 Err(err) => {
                     return Err(err);
