@@ -12,6 +12,8 @@
 //!
 //! # Examples
 //!
+//! We can store certificates in an in-memory store, and query the store:
+//!
 //! ```rust
 //! use std::sync::Arc;
 //! use sequoia_openpgp::cert::{Cert, CertBuilder};
@@ -43,6 +45,48 @@
 //! let cert: &Cert = cert.to_cert()?;
 //! # Ok(()) }
 //! ```
+//!
+//! We can use the same API on a persistent certificate store, kept on the
+//! filesystem:
+//!
+//! ```rust
+//! use std::fs;
+//! # use tempfile;
+//! #
+//! # use std::sync::Arc;
+//! #
+//! # use sequoia_openpgp as openpgp;
+//! # use openpgp::Result;
+//! # use openpgp::cert::{Cert, CertBuilder};
+//! #
+//! # use sequoia_cert_store::{CertStore, LazyCert, Store, StoreUpdate};
+//!
+//! # fn main() -> Result<()> {
+//! // Make a certificate store on the file system, in a fresh empty directory.
+//! let directory = "/tmp/test-sequoia-certificate-directory";
+//! # let directory_object = tempfile::tempdir()?; let directory = directory_object.path();
+//! fs::create_dir_all(directory)?;
+//! let mut store = CertStore::open(directory)?;
+//!
+//! // Make a new certificate.
+//! let (cert, _rev) = CertBuilder::new().generate()?;
+//! let fpr = cert.fingerprint();
+//!
+//! // The certificate of course will not be in the store yet.
+//! assert!(store.lookup_by_cert_fpr(&fpr).is_err());
+//!
+//! // Add it.
+//! store.update(Arc::new(LazyCert::from(cert)))?;
+//!
+//! // Now the certificate can be found.
+//! let found = store.lookup_by_cert_fpr(&fpr).expect("present");
+//! assert_eq!(found.fingerprint(), fpr);
+//!
+//! // Again, we can resolve the `LazyCert` to a `Cert`.
+//! let cert: &Cert = found.to_cert()?;
+//! # Ok(()) }
+//! ```
+
 use std::str;
 
 use sequoia_openpgp as openpgp;
@@ -173,7 +217,7 @@ mod tests {
 
     include!("../tests/keyring.rs");
 
-    fn test_backend<'a, B>(backend: B)
+    fn test_backend<'a, B>(backend: &B)
         where B: Store<'a>
     {
         // Check Store::list.
@@ -189,6 +233,8 @@ mod tests {
             assert_eq!(got, expected);
         }
 
+        std::thread::yield_now();
+
         // Check Store::iter.
         {
             let mut got: Vec<Fingerprint>
@@ -203,6 +249,8 @@ mod tests {
             assert_eq!(got, expected);
         }
 
+        std::thread::yield_now();
+
         // Iterate over the certificates in the keyring and check that
         // can look up the certificate by fingerprint, by key, by User
         // ID, and by email in various ways.
@@ -245,6 +293,8 @@ mod tests {
                         }
                     },
                 }
+
+                std::thread::yield_now();
             }
 
             // Check lookup_by_cert using key ids.
@@ -278,6 +328,8 @@ mod tests {
                         }
                     },
                 }
+
+                std::thread::yield_now();
             }
 
             // Check lookup_by_cert_or_subkey using fingerprints.
@@ -295,6 +347,8 @@ mod tests {
                 assert!(got.into_iter().any(|c| c.fingerprint() == fpr),
                         "{}, lookup_by_cert_or_subkey({}), with fingerprint, subkey",
                         handle.base, sk.fingerprint());
+
+                std::thread::yield_now();
             }
 
 
@@ -310,6 +364,8 @@ mod tests {
                     .expect("present");
                 assert!(got.into_iter().any(|c| c.fingerprint() == fpr),
                         "{}, lookup_by_cert_or_subkey, with keyid, subkey", handle.base);
+
+                std::thread::yield_now();
             }
 
 
@@ -527,6 +583,8 @@ mod tests {
                         }
                     }
                 }
+
+                std::thread::yield_now();
             }
         }
 
@@ -556,23 +614,53 @@ mod tests {
                         .parse::<Fingerprint>().expect("valid"),
             ]));
 
+        std::thread::yield_now();
 
         // ed's primary is also a subkey on the same certificate.
-        assert_eq!(
-            sort_vec(backend.lookup_by_cert_or_subkey(
-                &"0C346B2B6241263F64E9C7CF1EA300797258A74E"
-                    .parse::<KeyHandle>().expect("valid"))
-                .expect("present")
-                .into_iter()
-                .map(|c| c.fingerprint())
-                .collect::<Vec<Fingerprint>>()),
-            sort_vec(
-                vec![
-                    keyring::ed.fingerprint
-                        .parse::<Fingerprint>().expect("valid"),
-            ]));
+        {
+            let ed_fpr = "0C346B2B6241263F64E9C7CF1EA300797258A74E"
+                .parse::<Fingerprint>().expect("valid");
+            let ed_kh = KeyHandle::from(&ed_fpr);
 
+            assert_eq!(
+                sort_vec(backend.lookup_by_cert_or_subkey(&ed_kh)
+                         .expect("present")
+                         .into_iter()
+                         .map(|c| c.fingerprint())
+                         .collect::<Vec<Fingerprint>>()),
+                sort_vec(
+                    vec![
+                        keyring::ed.fingerprint
+                            .parse::<Fingerprint>().expect("valid"),
+                    ]));
 
+            // Also test that the CertStore implementation is not
+            // impacted.
+            let cert = backend.lookup_by_cert_fpr(&ed_fpr)
+                .expect("found cert");
+            assert_eq!(cert.fingerprint(), ed_fpr);
+
+            // Make sure the implementation doesn't return the certificate
+            // twice, once when matching on the primary key, and once when
+            // matching on the subkey.
+            let certs = backend.lookup_by_cert(&ed_kh)
+                .expect("found cert");
+            assert_eq!(certs.len(), 1);
+            assert_eq!(certs[0].fingerprint(), ed_fpr);
+
+            let certs = backend.lookup_by_cert_or_subkey(&ed_kh)
+                .expect("found cert");
+            assert_eq!(certs.len(), 1);
+            assert_eq!(certs[0].fingerprint(), ed_fpr);
+
+            let certs = backend.certs();
+            assert_eq!(certs.filter(|c| c.fingerprint() == ed_fpr).count(),
+                       1);
+
+            let fprs = backend.fingerprints();
+            assert_eq!(fprs.filter(|fpr| fpr == &ed_fpr).count(),
+                       1);
+        }
 
         // david has a subkey that doesn't have a binding signature,
         // but the backend is not supposed to check that.  (That
@@ -627,6 +715,8 @@ mod tests {
             }
         }
 
+        std::thread::yield_now();
+
         assert!(
             backend.lookup_by_cert_or_subkey(
                 &"0123 4567 89AB CDEF 0123 4567 89AB CDEF"
@@ -677,6 +767,8 @@ mod tests {
                   .parse::<Fingerprint>().expect("valid") ]);
 
 
+        std::thread::yield_now();
+
         // Check that when looking up a subdomain, we don't get back
         // User IDs in a subdomain.
         assert_eq!(
@@ -732,9 +824,6 @@ mod tests {
     fn certd() -> Result<()> {
         use std::io::Read;
 
-        // We expect 8 certificates.
-        assert_eq!(keyring::certs.len(), 12);
-
         let path = tempfile::tempdir()?;
         let certd = cert_d::CertD::with_base_dir(&path)
             .map_err(|err| {
@@ -763,7 +852,7 @@ mod tests {
         drop (certd);
 
         let certd = store::certd::CertD::open(&path).expect("exists");
-        test_backend(certd);
+        test_backend(&certd);
 
         Ok(())
     }
@@ -773,9 +862,6 @@ mod tests {
     fn cert_store() -> Result<()> {
         use std::io::Read;
 
-        // Sanity check how many certificates we read.
-        assert_eq!(keyring::certs.len(), 12);
-
         let path = tempfile::tempdir()?;
         let certd = cert_d::CertD::with_base_dir(&path)
             .map_err(|err| {
@@ -804,7 +890,7 @@ mod tests {
         drop(certd);
 
         let cert_store = CertStore::open(&path).expect("exists");
-        test_backend(cert_store);
+        test_backend(&cert_store);
 
         Ok(())
     }
@@ -813,8 +899,6 @@ mod tests {
     fn cert_store_layered() -> Result<()> {
         use std::io::Read;
 
-        assert_eq!(keyring::certs.len(), 12);
-
         // A certd for each certificate.
         let mut paths: Vec<tempfile::TempDir> = Vec::new();
 
@@ -852,7 +936,7 @@ mod tests {
             paths.push(path);
         }
 
-        test_backend(cert_store);
+        test_backend(&cert_store);
 
         Ok(())
     }
@@ -861,8 +945,6 @@ mod tests {
     fn certs() -> Result<()> {
         use std::io::Read;
 
-        assert_eq!(keyring::certs.len(), 12);
-
         let mut bytes = Vec::new();
         for cert in keyring::certs.iter() {
             let binary = cert.bytes();
@@ -875,7 +957,7 @@ mod tests {
 
         let backend = store::Certs::from_bytes(&bytes)
             .expect("valid");
-        test_backend(backend);
+        test_backend(&backend);
 
         Ok(())
     }
@@ -884,8 +966,6 @@ mod tests {
     fn certd_with_prefetch() -> Result<()> {
         use std::io::Read;
 
-        assert_eq!(keyring::certs.len(), 12);
-
         let path = tempfile::tempdir()?;
         let certd = cert_d::CertD::with_base_dir(&path)
             .map_err(|err| {
@@ -913,9 +993,9 @@ mod tests {
         }
         drop (certd);
 
-        let mut certd = store::CertD::open(&path).expect("exists");
+        let certd = store::CertD::open(&path).expect("exists");
         certd.prefetch_all();
-        test_backend(certd);
+        test_backend(&certd);
 
         Ok(())
     }
@@ -924,8 +1004,6 @@ mod tests {
     fn certs_with_prefetch() -> Result<()> {
         use std::io::Read;
 
-        assert_eq!(keyring::certs.len(), 12);
-
         let mut bytes = Vec::new();
         for cert in keyring::certs.iter() {
             let binary = cert.bytes();
@@ -936,10 +1014,10 @@ mod tests {
                 .expect(&format!("{}", cert.base));
         }
 
-        let mut backend = store::Certs::from_bytes(&bytes)
+        let backend = store::Certs::from_bytes(&bytes)
             .expect("valid");
         backend.prefetch_all();
-        test_backend(backend);
+        test_backend(&backend);
 
         Ok(())
     }
@@ -956,17 +1034,15 @@ mod tests {
                 PathBuf::from(&base).join(c.filename)
             }))?;
 
-        test_backend(cert_store);
+        test_backend(&cert_store);
 
         Ok(())
     }
 
     #[test]
-    fn pep() -> Result<()> {
+    fn certs_multithreaded() -> Result<()> {
         use std::io::Read;
 
-        assert_eq!(keyring::certs.len(), 12);
-
         let mut bytes = Vec::new();
         for cert in keyring::certs.iter() {
             let binary = cert.bytes();
@@ -977,9 +1053,47 @@ mod tests {
                 .expect(&format!("{}", cert.base));
         }
 
-        let mut backend = Pep::from_bytes(&bytes).expect("valid");
+        let backend = store::Certs::from_bytes(&bytes)
+            .expect("valid");
+
+        std::thread::scope(|s| {
+            let threads = (0..4usize)
+                .map(|_| {
+                    s.spawn(|| {
+                        // Really make sure all threads start by sleeping
+                        // for 10ms.
+                        std::thread::sleep(std::time::Duration::new(0, 10));
+
+                        test_backend(&backend);
+                    })
+                })
+                .collect::<Vec<_>>();
+
+            threads.into_iter().for_each(|h| {
+                h.join().expect("joined");
+            });
+        });
+
+        Ok(())
+    }
+
+    #[test]
+    fn pep() -> Result<()> {
+        use std::io::Read;
+
+        let mut bytes = Vec::new();
+        for cert in keyring::certs.iter() {
+            let binary = cert.bytes();
+            let mut reader = openpgp::armor::Reader::from_bytes(
+                &binary,
+                openpgp::armor::ReaderMode::VeryTolerant);
+            reader.read_to_end(&mut bytes)
+                .expect(&format!("{}", cert.base));
+        }
+
+        let backend = Pep::from_bytes(&bytes).expect("valid");
         backend.prefetch_all();
-        test_backend(backend);
+        test_backend(&backend);
 
         Ok(())
     }
@@ -987,7 +1101,7 @@ mod tests {
     // Make sure that when we update a certificate, we are able to
     // find any new components and we are still able to find the old
     // components.
-    fn test_store_update<'a, B>(mut backend: B) -> Result<()>
+    fn test_store_update<'a, B>(backend: B) -> Result<()>
         where B: store::StoreUpdate<'a>
     {
         let p = &StandardPolicy::new();
@@ -1126,5 +1240,46 @@ mod tests {
         let certs = Pep::empty()?;
         test_store_update(certs)
     }
+
+    #[test]
+    fn certs_multithreaded_update() -> Result<()> {
+        let mut certs = Vec::new();
+        for cert in keyring::certs.iter() {
+            let cert = Cert::from_bytes(&cert.bytes()).expect("valid");
+            certs.push(cert);
+        }
+
+        let mut fprs: Vec<Fingerprint>
+            = certs.iter().map(|cert| cert.fingerprint()).collect();
+        fprs.sort();
+        fprs.dedup();
+
+        let backend = store::Certs::empty();
+
+        std::thread::scope(|s| {
+            let backend = &backend;
+            let threads = certs.into_iter()
+                .map(|cert| {
+                    s.spawn(move || {
+                        // Really make sure all threads start by sleeping
+                        // for 10ms.
+                        std::thread::sleep(std::time::Duration::new(0, 10));
+
+                        backend.update(Arc::new(LazyCert::from(cert)))
+                            .expect("ok");
+                    })
+                })
+                .collect::<Vec<_>>();
+
+            threads.into_iter().for_each(|h| {
+                h.join().expect("joined");
+            });
+        });
+
+        assert_eq!(backend.certs().count(), fprs.len());
+
+        Ok(())
+    }
+
 }
 
