@@ -1,5 +1,7 @@
 use std::str;
 use std::sync::Arc;
+use std::sync::atomic::AtomicUsize;
+use std::sync::atomic::Ordering;
 
 use anyhow::Context;
 
@@ -39,6 +41,7 @@ pub struct UserIDQueryParams {
     email: bool,
     ignore_case: bool,
 }
+assert_send_and_sync!(UserIDQueryParams);
 
 impl UserIDQueryParams {
     /// Returns a new `UserIDQueryParams`.
@@ -552,7 +555,15 @@ pub trait Store<'a> {
 
         self.lookup_by_cert(&kh)
             .and_then(|v| {
-                assert!(v.len() <= 1);
+                assert!(v.len() <= 1,
+                        "Looking up {} returned multiple certificates: {}",
+                        fingerprint,
+                        v.iter()
+                            .map(|c| {
+                                c.fingerprint().to_string()
+                            })
+                            .collect::<Vec<String>>()
+                            .join(", "));
                 v.into_iter().next()
                     .ok_or(StoreError::NotFound(kh).into())
             })
@@ -697,7 +708,7 @@ pub trait Store<'a> {
     ///
     /// Errors should be silently ignored and propagated when the
     /// operation in question is executed directly.
-    fn prefetch_all(&mut self) {
+    fn prefetch_all(&self) {
     }
 
     /// Prefetches some certificates.
@@ -711,7 +722,7 @@ pub trait Store<'a> {
     ///
     /// Errors should be silently ignored and propagated when the
     /// operation in question is executed directly.
-    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
+    fn prefetch_some(&self, certs: &[KeyHandle]) {
         let _ = certs;
     }
 }
@@ -773,11 +784,11 @@ where T: Store<'a> + ?Sized + 't
         self.as_ref().certs()
     }
 
-    fn prefetch_all(&mut self) {
+    fn prefetch_all(&self) {
         self.as_ref().prefetch_all()
     }
 
-    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
+    fn prefetch_some(&self, certs: &[KeyHandle]) {
         self.as_ref().prefetch_some(certs)
     }
 }
@@ -891,11 +902,11 @@ where T: Store<'a> + ?Sized
         (**self).certs()
     }
 
-    fn prefetch_all(&mut self) {
+    fn prefetch_all(&self) {
         (**self).prefetch_all()
     }
 
-    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
+    fn prefetch_some(&self, certs: &[KeyHandle]) {
         (**self).prefetch_some(certs)
     }
 }
@@ -914,7 +925,12 @@ pub trait MergeCerts<'a> {
     /// [`Cert::merge_public`].  This means that any secret key
     /// material in `disk` is preserved, any secret key material in
     /// `new` is ignored, and unhashed subpacket areas are merged.
-    fn merge_public<'b>(&mut self,
+    ///
+    /// Note: `self` is a normal reference, not a mutable reference.
+    /// This means that the underlying implementation has to use
+    /// interior mutability.  The advantage is that multiple threads
+    /// can have a reference to the store, and update it in parallel.
+    fn merge_public<'b>(&self,
                         new: Arc<LazyCert<'a>>,
                         disk: Option<Arc<LazyCert<'b>>>)
         -> Result<Arc<LazyCert<'a>>>
@@ -943,7 +959,12 @@ pub trait StoreUpdate<'a>: Store<'a> {
     ///
     /// This uses the default implementation of [`MergeCerts`] to
     /// merge the certificate with any existing certificate.
-    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
+    ///
+    /// Note: `self` is a normal reference, not a mutable reference.
+    /// This means that the underlying implementation has to use
+    /// interior mutability.  The advantage is that multiple threads
+    /// can have a reference to the store, and update it in parallel.
+    fn update(&self, cert: Arc<LazyCert<'a>>) -> Result<()> {
         self.update_by(cert, &mut ())?;
 
         Ok(())
@@ -963,35 +984,41 @@ pub trait StoreUpdate<'a>: Store<'a> {
     ///
     /// To use the default merge strategy, either call
     /// [`StoreUpdate::update`] directly, or pass `&mut ()`.
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                 merge_strategy: &mut dyn MergeCerts<'a>)
+    ///
+    /// Note: `self` and `merge_strategy` are normal references, not a
+    /// mutable reference.  This means that the underlying
+    /// implementation has to use interior mutability.  The advantage
+    /// is that multiple threads can have a reference to the store,
+    /// and update it in parallel.
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>;
 }
 
 impl<'a: 't, 't, T> StoreUpdate<'a> for Box<T>
 where T: StoreUpdate<'a> + ?Sized + 't
 {
-    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
-        self.as_mut().update(cert)
+    fn update(&self, cert: Arc<LazyCert<'a>>) -> Result<()> {
+        self.as_ref().update(cert)
     }
 
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a>)
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>
     {
-        self.as_mut().update_by(cert, merge_strategy)
+        self.as_ref().update_by(cert, merge_strategy)
     }
 }
 
-impl<'a: 't, 't, T> StoreUpdate<'a> for &'t mut T
+impl<'a: 't, 't, T> StoreUpdate<'a> for &'t T
 where T: StoreUpdate<'a> + ?Sized
 {
-    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
+    fn update(&self, cert: Arc<LazyCert<'a>>) -> Result<()> {
         (*self).update(cert)
     }
 
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a>)
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>
     {
         (*self).update_by(cert, merge_strategy)
@@ -1002,36 +1029,77 @@ where T: StoreUpdate<'a> + ?Sized
 ///
 /// This is primarily useful as the `merge_strategy` callback to
 /// [`StoreUpdate::update_by`].
-#[derive(Debug, Clone)]
+#[derive(Debug)]
 #[non_exhaustive]
 pub struct MergePublicCollectStats {
     /// Number of new certificates.
-    pub new: usize,
+    new_certs: AtomicUsize,
 
     /// Number of unchanged certificates.
     ///
     /// Note: there may be false negative.  That is some certificates
     /// may be unchanged, but the heuristic thinks that they have been
     /// updated.
-    pub unchanged: usize,
+    unchanged_certs: AtomicUsize,
 
     /// Number of update certificates.
-    pub updated: usize,
+    updated_certs: AtomicUsize,
 
     /// Number of errors.
-    pub errors: usize,
+    errors: AtomicUsize,
 }
+assert_send_and_sync!(MergePublicCollectStats);
 
 impl MergePublicCollectStats {
     /// Returns a new `MergePublicCollectStats` with all stats set to 0.
     pub fn new() -> Self {
         Self {
-            new: 0,
-            unchanged: 0,
-            updated: 0,
-            errors: 0,
+            new_certs: AtomicUsize::new(0),
+            unchanged_certs: AtomicUsize::new(0),
+            updated_certs: AtomicUsize::new(0),
+            errors: AtomicUsize::new(0),
         }
     }
+
+    /// Returns the number of new certificates.
+    pub fn new_certs(&self) -> usize {
+        self.new_certs.load(Ordering::Relaxed)
+    }
+
+    /// Returns the number of unchanged certificates.
+    pub fn unchanged_certs(&self) -> usize {
+        self.unchanged_certs.load(Ordering::Relaxed)
+    }
+
+    /// Returns the number of updated certificates.
+    pub fn updated_certs(&self) -> usize {
+        self.updated_certs.load(Ordering::Relaxed)
+    }
+
+    /// Returns the number of errors.
+    pub fn errors(&self) -> usize {
+        self.errors.load(Ordering::Relaxed)
+    }
+
+    /// Increments the number of new certificates by 1.
+    pub fn inc_new_certs(&self) {
+        self.new_certs.fetch_add(1, Ordering::Relaxed);
+    }
+
+    /// Increments the number of unchanged certificates by 1.
+    pub fn inc_unchanged_certs(&self) {
+        self.unchanged_certs.fetch_add(1, Ordering::Relaxed);
+    }
+
+    /// Increments the number of updated certificates by 1.
+    pub fn inc_updated_certs(&self) {
+        self.updated_certs.fetch_add(1, Ordering::Relaxed);
+    }
+
+    /// Increments the number of errors by 1.
+    pub fn inc_errors(&self) {
+        self.errors.fetch_add(1, Ordering::Relaxed);
+    }
 }
 
 impl<'a> MergeCerts<'a> for MergePublicCollectStats {
@@ -1071,10 +1139,10 @@ impl<'a> MergeCerts<'a> for MergePublicCollectStats {
     /// certs.update_by(Arc::new(LazyCert::from(cert)), &mut stats)
     ///         .expect("valid");
     ///
-    /// assert_eq!(stats.new, 1);
+    /// assert_eq!(stats.new_certs(), 1);
     /// # Ok(()) }
     /// ```
-    fn merge_public<'b, 'rb>(&mut self,
+    fn merge_public<'b, 'rb>(&self,
                              new: Arc<LazyCert<'a>>,
                              disk: Option<Arc<LazyCert<'b>>>)
         -> Result<Arc<LazyCert<'a>>>
@@ -1082,7 +1150,7 @@ impl<'a> MergeCerts<'a> for MergePublicCollectStats {
         let disk = if let Some(disk) = disk {
             disk
         } else {
-            self.new += 1;
+            self.inc_new_certs();
 
             if new.is_tsk() {
                 return Ok(Arc::new(LazyCert::from(
@@ -1109,18 +1177,22 @@ impl<'a> MergeCerts<'a> for MergePublicCollectStats {
             .clone();
 
         if disk == new {
-            self.unchanged += 1;
+            self.inc_unchanged_certs();
             Ok(Arc::new(LazyCert::from(new)))
         } else {
             // If the on-disk version has secrets, we preserve them.
             // If new has secrets, we ignore them.
-            match disk.merge_public(new) {
+            match disk.clone().merge_public(new) {
                 Ok(merged) => {
-                    self.updated += 1;
+                    if merged == disk {
+                        self.inc_unchanged_certs();
+                    } else {
+                        self.inc_updated_certs();
+                    }
                     Ok(Arc::new(LazyCert::from(merged)))
                 }
                 Err(err) => {
-                    self.errors += 1;
+                    self.inc_errors();
                     Err(err.into())
                 }
             }
@@ -1199,8 +1271,8 @@ mod tests {
         }
 
         let backend = store::Certs::empty();
-        let mut backend: Box<dyn StoreUpdate> = Box::new(backend);
-        let foo = Foo::new(&mut backend);
+        let backend: Box<dyn StoreUpdate> = Box::new(backend);
+        let foo = Foo::new(&backend);
 
         // Do something (anything) with the backend.
         assert_eq!(foo.count(), 0);
@@ -1258,9 +1330,7 @@ mod tests {
         use crate::CertStore;
         use crate::store::MergePublicCollectStats;
 
-        assert_eq!(keyring::certs.len(), 12);
-
-        let mut certs = CertStore::empty();
+        let certs = CertStore::empty();
 
         let mut stats = MergePublicCollectStats::new();
 
@@ -1277,13 +1347,16 @@ mod tests {
             eprintln!("After inserting {} ({}), stats: {:?}",
                       i, fpr, stats);
 
-            assert_eq!(stats.new, seen.len());
-            assert_eq!(stats.new + stats.updated + stats.unchanged, i + 1);
+            assert_eq!(stats.new_certs(), seen.len());
+            assert_eq!(stats.new_certs()
+                       + stats.updated_certs()
+                       + stats.unchanged_certs(),
+                       i + 1);
         }
 
-        let new = stats.new;
-        let updated = stats.updated;
-        let unchanged = stats.unchanged;
+        let new = stats.new_certs();
+        let updated = stats.updated_certs();
+        let unchanged = stats.unchanged_certs();
 
         // Insert again.  This time nothing should change.
         for (i, cert) in keyring::certs.iter().enumerate() {
@@ -1297,10 +1370,10 @@ mod tests {
                       i, fpr, stats);
 
             // These should not change:
-            assert_eq!(stats.new, new);
+            assert_eq!(stats.new_certs(), new);
             // Update should also not change, but there may be false
             // positives.
-            assert_eq!(stats.unchanged + stats.updated,
+            assert_eq!(stats.unchanged_certs() + stats.updated_certs(),
                        updated + unchanged + i + 1);
         }
     }
