@@ -1,6 +1,7 @@
 use std::collections::BTreeMap;
 use std::collections::btree_map::Entry;
 use std::sync::Arc;
+use std::sync::RwLock;
 
 use smallvec::SmallVec;
 use smallvec::smallvec;
@@ -14,6 +15,7 @@ use openpgp::cert::raw::RawCertParser;
 use openpgp::Fingerprint;
 use openpgp::KeyID;
 use openpgp::KeyHandle;
+use openpgp::packet::UserID;
 use openpgp::parse::Parse;
 use openpgp::Result;
 
@@ -32,6 +34,11 @@ const TRACE: bool = false;
 /// `Cert`s implements `StoreUpdate`, but it does not write the
 /// updates to disk; they are only updated in memory.
 pub struct Certs<'a> {
+    inner: RwLock<CertsInner<'a>>,
+}
+assert_send_and_sync!(Certs<'_>);
+
+struct CertsInner<'a> {
     // Indexed by primary key fingerprint.
     certs: BTreeMap<Fingerprint, Arc<LazyCert<'a>>>,
     // Indexed by a key's KeyID (primary key or subkey) and maps to
@@ -49,9 +56,11 @@ impl<'a> Certs<'a>
     /// be added to it using the [`StoreUpdate`] interface.
     pub fn empty() -> Self {
         Certs {
-            certs: Default::default(),
-            keys: Default::default(),
-            userid_index: UserIDIndex::new(),
+            inner: RwLock::new(CertsInner {
+                certs: Default::default(),
+                keys: Default::default(),
+                userid_index: UserIDIndex::new(),
+            }),
         }
     }
 
@@ -77,7 +86,7 @@ impl<'a> Certs<'a>
         -> Result<Self>
         where I: Into<LazyCert<'a>>
     {
-        let mut r = Self::empty();
+        let r = Self::empty();
         for cert in certs {
             r.update(Arc::new(cert.into())).expect("implementation doesn't fail")
         }
@@ -86,7 +95,7 @@ impl<'a> Certs<'a>
     }
 }
 
-impl<'a> Store<'a> for Certs<'a>
+impl<'a> Store<'a> for CertsInner<'a>
 {
     fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
         tracer!(TRACE, "Certs::lookup_by_cert");
@@ -182,16 +191,101 @@ impl<'a> Store<'a> for Certs<'a>
         Box::new(self.certs.values().cloned())
     }
 
-    fn prefetch_all(&mut self) {
-        self.prefetch(true, &[])
+    fn prefetch_all(&self) {
+        // Because the interior mutability is implemented via Certs,
+        // not CertInner, and we need a mutable reference here, but we
+        // only have a normal, non-mutable reference, the prefetch
+        // implementation is dispatched via `Certs`.
     }
 
-    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
-        self.prefetch(false, certs)
+    fn prefetch_some(&self, _certs: &[KeyHandle]) {
+        // Because the interior mutability is implemented via Certs,
+        // not CertInner, and we need a mutable reference here, but we
+        // only have a normal, non-mutable reference, the prefetch
+        // implementation is dispatched via `Certs`.
     }
 }
 
-impl Certs<'_> {
+impl<'a> Store<'a> for Certs<'a>
+{
+    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.lookup_by_cert(kh)
+    }
+
+    fn lookup_by_cert_fpr(&self, fingerprint: &Fingerprint)
+        -> Result<Arc<LazyCert<'a>>>
+    {
+        let inner = self.inner.read().unwrap();
+        inner.lookup_by_cert_fpr(fingerprint)
+    }
+
+    fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.lookup_by_cert_or_subkey(kh)
+    }
+
+    fn select_userid(&self, query: &UserIDQueryParams, pattern: &str)
+        -> Result<Vec<Arc<LazyCert<'a>>>>
+    {
+        let inner = self.inner.read().unwrap();
+        inner.select_userid(query, pattern)
+    }
+
+    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.lookup_by_userid(userid)
+    }
+
+    fn grep_userid(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.grep_userid(pattern)
+    }
+
+    fn lookup_by_email(&self, email: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.lookup_by_email(email)
+    }
+
+    fn grep_email(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.grep_email(pattern)
+    }
+
+    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        let inner = self.inner.read().unwrap();
+        inner.lookup_by_email_domain(domain)
+    }
+
+    fn fingerprints<'b>(&'b self) -> Box<dyn Iterator<Item=Fingerprint> + 'b> {
+        // We aren't lazy to avoid holding the lock.
+        let inner = self.inner.read().unwrap();
+        let fprs: Vec<_> = inner.fingerprints().collect();
+        Box::new(fprs.into_iter())
+    }
+
+    fn certs<'b>(&'b self)
+        -> Box<dyn Iterator<Item=Arc<LazyCert<'a>>> + 'b>
+        where 'a: 'b
+    {
+        // We aren't lazy to avoid holding the lock.
+        let inner = self.inner.read().unwrap();
+        let certs: Vec<_> = inner.certs().collect();
+        Box::new(certs.into_iter())
+    }
+
+    fn prefetch_all(&self) {
+        let mut inner = self.inner.write().unwrap();
+        inner.prefetch(true, &[])
+    }
+
+    fn prefetch_some(&self, certs: &[KeyHandle]) {
+        let mut inner = self.inner.write().unwrap();
+        inner.prefetch(false, certs)
+    }
+}
+
+impl CertsInner<'_> {
     /// Prefetches the certs identified by `khs`, or `all`.
     fn prefetch(&mut self, all: bool, khs: &[KeyHandle]) {
         // LazyCert is Sync and Send, but keeping a reference to the
@@ -325,17 +419,19 @@ impl Certs<'_> {
 }
 
 impl<'a> StoreUpdate<'a> for Certs<'a> {
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                 merge_strategy: &mut dyn MergeCerts<'a>)
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>
     {
         tracer!(TRACE, "Certs::update_by");
 
+        let mut inner = self.inner.write().unwrap();
+
         let fpr = cert.fingerprint();
 
         // Add the cert fingerprint -> cert entry.
         let merged: Arc<LazyCert>;
-        match self.certs.entry(fpr.clone()) {
+        match inner.certs.entry(fpr.clone()) {
             Entry::Occupied(mut oe) => {
                 t!("Updating {}", fpr);
 
@@ -359,7 +455,7 @@ impl<'a> StoreUpdate<'a> for Certs<'a> {
         // Populate the key map.  This is a merge so we are not
         // removing anything.
         for k in merged.keys() {
-            match self.keys.entry(k.keyid()) {
+            match inner.keys.entry(k.keyid()) {
                 Entry::Occupied(mut oe) => {
                     let fprs = oe.get_mut();
                     if ! fprs.contains(&fpr) {
@@ -372,7 +468,7 @@ impl<'a> StoreUpdate<'a> for Certs<'a> {
             }
         }
 
-        self.userid_index.insert(&fpr, merged.userids());
+        inner.userid_index.insert(&fpr, merged.userids());
 
         Ok(merged)
     }
