@@ -1,7 +1,7 @@
-use std::cell::RefCell;
 use std::collections::BTreeMap;
 use std::collections::BTreeSet;
 use std::sync::Arc;
+use std::sync::Mutex;
 
 use sequoia_openpgp as openpgp;
 use openpgp::Cert;
@@ -38,40 +38,47 @@ pub const PROTON_URL: &str = "hkps://api.protonmail.ch";
 
 /// A keyserver backend.
 pub struct KeyServer<'a> {
-    keyserver: RefCell<net::KeyServer>,
+    inner: Mutex<KeyServerInner<'a>>,
+}
+assert_send_and_sync!(KeyServer<'_>);
+
+struct KeyServerInner<'a> {
+    keyserver: net::KeyServer,
 
     id: String,
-    tx: RefCell<usize>,
-    listeners: Vec<Box<dyn StatusListener>>,
+    tx: usize,
+    listeners: Vec<Box<dyn StatusListener + Send + Sync>>,
 
     // A cache.  We only cache certificates; we don't cache User ID
     // searches.
 
     // Primary keys and subkeys.
-    hits_fpr: RefCell<BTreeMap<Fingerprint, Arc<LazyCert<'a>>>>,
-    hits_keyid: RefCell<BTreeMap<KeyID, Fingerprint>>,
+    hits_fpr: BTreeMap<Fingerprint, Arc<LazyCert<'a>>>,
+    hits_keyid: BTreeMap<KeyID, Fingerprint>,
     // What we failed to look up.
-    misses_fpr: RefCell<BTreeSet<Fingerprint>>,
-    misses_keyid: RefCell<BTreeSet<KeyID>>,
+    misses_fpr: BTreeSet<Fingerprint>,
+    misses_keyid: BTreeSet<KeyID>,
 }
 
 impl KeyServer<'_> {
     /// Returns a new key server instance.
     pub fn new(url: &str) -> Result<Self> {
         Ok(Self {
-            keyserver: RefCell::new(net::KeyServer::new(url)?),
-            id: if url.len() <= 10 {
-                // Only prefix "key server" if the URL is short.
-                format!("key server {}", url)
-            } else {
-                url.to_string()
-            },
-            tx: RefCell::new(0),
-            listeners: Vec::new(),
-            hits_fpr: Default::default(),
-            hits_keyid: Default::default(),
-            misses_fpr: Default::default(),
-            misses_keyid: Default::default(),
+            inner: Mutex::new(KeyServerInner {
+                keyserver: net::KeyServer::new(url)?,
+                id: if url.len() <= 10 {
+                    // Only prefix "key server" if the URL is short.
+                    format!("key server {}", url)
+                } else {
+                    url.to_string()
+                },
+                tx: 0,
+                listeners: Vec::new(),
+                hits_fpr: Default::default(),
+                hits_keyid: Default::default(),
+                misses_fpr: Default::default(),
+                misses_keyid: Default::default(),
+            }),
         })
     }
 
@@ -98,22 +105,22 @@ impl KeyServer<'_> {
     }
 
     /// Sends status updates to the listener.
-    pub fn add_listener(&mut self, listener: Box<dyn StatusListener>) {
-        self.listeners.push(listener);
+    pub fn add_listener(&mut self, listener: Box<dyn StatusListener + Send + Sync>) {
+        self.inner.lock().unwrap().listeners.push(listener);
     }
 }
 
-impl<'a> KeyServer<'a> {
+impl<'a> KeyServerInner<'a> {
     // Looks for a certificate in the cache.
     fn check_cache(&self, kh: &KeyHandle)
         -> Option<Result<Vec<Arc<LazyCert<'a>>>>>
     {
         let kh_;
         let kh = if let KeyHandle::KeyID(keyid) = kh {
-            if let Some(fpr) = self.hits_keyid.borrow().get(keyid) {
+            if let Some(fpr) = self.hits_keyid.get(keyid) {
                 kh_ = KeyHandle::Fingerprint(fpr.clone());
                 &kh_
-            } else if self.misses_keyid.borrow().get(keyid).is_some() {
+            } else if self.misses_keyid.get(keyid).is_some() {
                 return Some(Err(StoreError::NotFound(
                     KeyHandle::from(kh.clone())).into()));
             } else {
@@ -123,10 +130,10 @@ impl<'a> KeyServer<'a> {
             kh
         };
         if let KeyHandle::Fingerprint(fpr) = kh {
-            if let Some(cert) = self.hits_fpr.borrow().get(fpr) {
+            if let Some(cert) = self.hits_fpr.get(fpr) {
                 return Some(Ok(vec![ cert.clone() ]));
             }
-            if self.misses_fpr.borrow().get(fpr).is_some() {
+            if self.misses_fpr.get(fpr).is_some() {
                 return Some(Err(StoreError::NotFound(
                     KeyHandle::from(kh.clone())).into()));
             }
@@ -136,24 +143,18 @@ impl<'a> KeyServer<'a> {
     }
 
     // Adds the cert to the in-memory cache.
-    fn cache(&self, cert: &Arc<LazyCert<'a>>) {
-        let mut hits_fpr = self.hits_fpr.borrow_mut();
-        let mut hits_keyid = self.hits_keyid.borrow_mut();
-
+    fn cache(&mut self, cert: &Arc<LazyCert<'a>>) {
         for k in cert.keys() {
-            hits_fpr.insert(k.fingerprint(), cert.clone());
-            hits_keyid.insert(k.keyid(), k.fingerprint());
+            self.hits_fpr.insert(k.fingerprint(), cert.clone());
+            self.hits_keyid.insert(k.keyid(), k.fingerprint());
         }
     }
 
     /// Deletes the key with the given fingerprint from the cache.
-    pub(crate) fn delete_from_cache(&mut self, cert: &LazyCert<'a>) {
-        let mut hits_fpr = self.hits_fpr.borrow_mut();
-        let mut hits_keyid = self.hits_keyid.borrow_mut();
-
+    fn delete_from_cache(&mut self, cert: &LazyCert<'_>) {
         for key in cert.keys() {
             let fp = key.fingerprint();
-            hits_fpr.remove(&fp);
+            self.hits_fpr.remove(&fp);
 
             let keyid = KeyID::from(fp);
             // XXX: Technically, a key ID could map to multiple
@@ -161,7 +162,19 @@ impl<'a> KeyServer<'a> {
             // But, by using a BTreeMap<KeyID, Fingerprint> we do accept
             // this kind of loss in the first place, as later cache
             // insertions will overwrite any existing mappings.
-            hits_keyid.remove(&keyid);
+            self.hits_keyid.remove(&keyid);
+        }
+    }
+}
+
+impl<'a> KeyServer<'a> {
+    /// Deletes the key with the given fingerprint from the cache.
+    pub(crate) fn delete_from_cache<I>(&self, certs: I)
+        where I: Iterator<Item=Arc<LazyCert<'a>>>,
+    {
+        let mut inner = self.inner.lock().unwrap();
+        for cert in certs {
+            inner.delete_from_cache(&cert);
         }
     }
 }
@@ -171,7 +184,7 @@ macro_rules! update {
     ( $self:expr, $update:ident, $($args:expr),* ) => {{
         if ! $self.listeners.is_empty() {
             let update = StatusUpdate::$update(
-                *$self.tx.borrow(), &$self.id, $($args),*);
+                $self.tx, &$self.id, $($args),*);
 
             for sub in $self.listeners.iter() {
                 sub.update(&update);
@@ -202,21 +215,21 @@ impl<'a> Store<'a> for KeyServer<'a> {
 
         // Check the cache.
         t!("Looking up {} on keyserver...", kh);
-        if ! self.listeners.is_empty() {
-            let tx = *self.tx.borrow();
-            *self.tx.borrow_mut() = tx + 1;
-            update!(self, LookupStarted, kh, None);
+        let mut inner = self.inner.lock().unwrap();
+        if ! inner.listeners.is_empty() {
+            inner.tx += 1;
+            update!(inner, LookupStarted, kh, None);
         };
 
-        if let Some(r) = self.check_cache(kh) {
+        if let Some(r) = inner.check_cache(kh) {
             t!("Found in in-memory cache");
             match r.as_ref() {
                 Ok(certs) => {
-                    update!(self, LookupFinished, kh,
+                    update!(inner, LookupFinished, kh,
                             &certs[..], Some("Found in in-memory cache"));
                 }
                 Err(err) => {
-                    update!(self, LookupFailed, kh, Some(&err));
+                    update!(inner, LookupFailed, kh, Some(&err));
                 }
             }
             return r;
@@ -228,7 +241,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
         // The keyserver interface currently only returns a single
         // result.
         let r = rt.block_on(async {
-            self.keyserver.borrow_mut().get(kh.clone()).await
+            inner.keyserver.get(kh.clone()).await
         });
 
         match r {
@@ -242,7 +255,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
                     t!("keyserver returned {}", cert.fingerprint());
 
                     // Add the result to the cache.
-                    self.cache(cert);
+                    inner.cache(cert);
                 }
 
                 // Make sure the key server gave us the right
@@ -254,7 +267,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
                     .collect();
 
                 if ! requested.is_empty() {
-                    update!(self, LookupFinished, kh, &requested[..], None);
+                    update!(inner, LookupFinished, kh, &requested[..], None);
                     Ok(requested)
                 } else {
                     let err = KeyServerError::UnexpectedResults(
@@ -262,7 +275,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
                         certs.iter().map(|c| c.fingerprint()).collect())
                         .into();
                     t!("{}", err);
-                    update!(self, LookupFailed, kh, Some(&err));
+                    update!(inner, LookupFailed, kh, Some(&err));
                     Err(StoreError::NotFound(
                         KeyHandle::from(kh.clone())).into())
                 }
@@ -273,11 +286,11 @@ impl<'a> Store<'a> for KeyServer<'a> {
                 if let Some(net::Error::NotFound)
                     = err.downcast_ref::<net::Error>()
                 {
-                    update!(self, LookupFailed, kh, None);
+                    update!(inner, LookupFailed, kh, None);
                     Err(StoreError::NotFound(
                         KeyHandle::from(kh.clone())).into())
                 } else {
-                    update!(self, LookupFailed, kh, Some(&err));
+                    update!(inner, LookupFailed, kh, Some(&err));
                     Err(err)
                 }
             }
@@ -291,10 +304,10 @@ impl<'a> Store<'a> for KeyServer<'a> {
 
         t!("{}", pattern);
         t!("Looking {:?} up on the keyserver... ", pattern);
-        if ! self.listeners.is_empty() {
-            let tx = *self.tx.borrow();
-            *self.tx.borrow_mut() = tx + 1;
-            update!(self, SearchStarted, pattern, None);
+        let mut inner = self.inner.lock().unwrap();
+        if ! inner.listeners.is_empty() {
+            inner.tx += 1;
+            update!(inner, SearchStarted, pattern, None);
         }
 
         let email = if query.email && query.anchor_start && query.anchor_end {
@@ -318,8 +331,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
         let rt = tokio::runtime::Runtime::new().unwrap();
         let (ks, wkd) = rt.block_on(async {
             // Query the keyserver.
-            let ks = self.keyserver.borrow_mut();
-            let ks = ks.search(pattern);
+            let ks = inner.keyserver.search(pattern);
 
             // And the WKD (if it appears to be an email address).
             let wkd = async {
@@ -338,9 +350,9 @@ impl<'a> Store<'a> for KeyServer<'a> {
         match ks {
             Ok(c) => {
                 t!("Key server returned {} results", c.len());
-                if ! self.listeners.is_empty() {
+                if ! inner.listeners.is_empty() {
                     let msg = format!("Key server returned {} results", c.len());
-                    update!(self, SearchStatus, pattern, &msg);
+                    update!(inner, SearchStatus, pattern, &msg);
                 }
                 certs.extend(c);
             },
@@ -349,10 +361,10 @@ impl<'a> Store<'a> for KeyServer<'a> {
         match wkd {
             Ok(c) => {
                 t!("WKD server returned {} results", c.len());
-                if ! self.listeners.is_empty() {
+                if ! inner.listeners.is_empty() {
                     let msg = format!("WKD server returned {} results",
                                       c.len());
-                    update!(self, SearchStatus, pattern, &msg);
+                    update!(inner, SearchStatus, pattern, &msg);
                 }
                 certs.extend(c);
             },
@@ -383,7 +395,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
             certs.into_iter().map(|c| Arc::new(LazyCert::from(c))).collect();
 
         // Add the results to the cache.
-        certs.iter().for_each(|cert| self.cache(cert));
+        certs.iter().for_each(|cert| inner.cache(cert));
 
         // Only keep the certificates that actually satisfy the
         // constraints.
@@ -394,10 +406,10 @@ impl<'a> Store<'a> for KeyServer<'a> {
         });
 
         if certs.is_empty() {
-            update!(self, SearchFailed, pattern, None);
+            update!(inner, SearchFailed, pattern, None);
             Err(StoreError::NoMatches(pattern.to_string()).into())
         } else {
-            if TRACE || ! self.listeners.is_empty() {
+            if TRACE || ! inner.listeners.is_empty() {
                 let msg = format!(
                     "Got {} results:\n  {}",
                     certs.len(),
@@ -424,7 +436,7 @@ impl<'a> Store<'a> for KeyServer<'a> {
                         .collect::<Vec<_>>()
                         .join("\n  "));
                 t!("{}", msg);
-                update!(self, SearchFinished, pattern, &certs[..], Some(&msg));
+                update!(inner, SearchFinished, pattern, &certs[..], Some(&msg));
             }
             Ok(certs)
         }
@@ -433,8 +445,8 @@ impl<'a> Store<'a> for KeyServer<'a> {
     fn fingerprints<'b>(&'b self) -> Box<dyn Iterator<Item=Fingerprint> + 'b> {
         // Return the entries in our cache.
         Box::new(self
+                 .inner.lock().unwrap()
                  .hits_fpr
-                 .borrow()
                  .keys()
                  .cloned()
                  .collect::<Vec<_>>()
