@@ -1,6 +1,7 @@
 use std::path::Path;
 use std::path::PathBuf;
 use std::sync::Arc;
+use std::sync::Mutex;
 
 use anyhow::Context;
 
@@ -75,8 +76,9 @@ macro_rules! wrap_err {
 ///
 /// [pEp]: https://gitea.pep.foundation/pEp.foundation/pEpEngine
 pub struct Pep {
-    conn: rusqlite::Connection,
+    conn: Mutex<rusqlite::Connection>,
 }
+assert_send_and_sync!(Pep);
 
 // Generates a convenience method that returns a prepared statement
 // for the specified sql.  If preparing the statement results in an
@@ -204,7 +206,7 @@ impl Pep {
         -> Result<Self>
         where I: Into<LazyCert<'a>>
     {
-        let mut r = Self::init_in_memory()?;
+        let r = Self::init_in_memory()?;
         for cert in certs {
             r.update(Arc::new(cert.into())).expect("implementation doesn't fail")
         }
@@ -314,7 +316,7 @@ impl Pep {
                     keys_db.display()))?;
 
         Ok(Pep {
-            conn,
+            conn: Mutex::new(conn),
         })
     }
 
@@ -445,7 +447,8 @@ impl Pep {
     {
         tracer!(TRACE, "Pep::tsk_lookup_by_cert_fpr");
 
-        let mut stmt = Self::tsk_find_stmt(&self.conn)?;
+        let conn = self.conn.lock().unwrap();
+        let mut stmt = Self::tsk_find_stmt(&conn)?;
 
         let r = cert_query!(stmt, [ fpr.to_hex() ],
                             StoreError::NotFound(KeyHandle::from(fpr)))?;
@@ -464,7 +467,8 @@ impl Pep {
     {
         tracer!(TRACE, "Pep::tsk_lookup_by_cert_or_subkey");
 
-        let mut stmt = Self::tsk_find_with_key_stmt(&self.conn)?;
+        let conn = self.conn.lock().unwrap();
+        let mut stmt = Self::tsk_find_with_key_stmt(&conn)?;
 
         let keyid = KeyID::from(kh).to_hex();
         t!("({})", keyid);
@@ -482,7 +486,8 @@ impl Pep {
         tracer!(TRACE, "Pep::tsks");
 
         let inner = || -> Result<Vec<_>> {
-            let mut stmt = Self::tsk_all_stmt(&self.conn)?;
+            let conn = self.conn.lock().unwrap();
+            let mut stmt = Self::tsk_all_stmt(&conn)?;
             cert_query!(stmt, [ ], StoreError::NoMatches("EOF".into()))
         };
 
@@ -502,8 +507,10 @@ impl Pep {
     ///
     /// Returns an error if the specified certificate is not found.
     pub fn cert_delete(&mut self, fpr: Fingerprint) -> Result<()> {
+        let conn = self.conn.lock().unwrap();
+
         let changes = wrap_err!(
-            Self::cert_delete_stmt(&self.conn)?
+            Self::cert_delete_stmt(&conn)?
                 .execute(params![ fpr.to_hex() ]),
             CannotDeleteKey,
             format!("Deleting {}", fpr))?;
@@ -526,12 +533,14 @@ impl<'a> Store<'a> for Pep {
     fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
         tracer!(TRACE, "Pep::lookup_by_cert");
 
+        let conn = self.conn.lock().unwrap();
+
         let mut stmt = match kh {
             KeyHandle::Fingerprint(_) => {
-                Self::cert_find_stmt(&self.conn)?
+                Self::cert_find_stmt(&conn)?
             }
             KeyHandle::KeyID(_) => {
-                Self::cert_find_by_keyid_stmt(&self.conn)?
+                Self::cert_find_by_keyid_stmt(&conn)?
             }
         };
 
@@ -549,7 +558,9 @@ impl<'a> Store<'a> for Pep {
     fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
         tracer!(TRACE, "Pep::lookup_by_cert_or_subkey");
 
-        let mut stmt = Self::cert_find_with_key_stmt(&self.conn)?;
+        let conn = self.conn.lock().unwrap();
+
+        let mut stmt = Self::cert_find_with_key_stmt(&conn)?;
 
         let keyid = KeyID::from(kh).to_hex();
         t!("({})", keyid);
@@ -630,7 +641,9 @@ impl<'a> Store<'a> for Pep {
         let userid = crate::email_to_userid(&email)?;
         let email = userid.email_normalized()?.expect("have one");
 
-        let mut stmt = Self::cert_find_by_email_stmt(&self.conn)?;
+        let conn = self.conn.lock().unwrap();
+
+        let mut stmt = Self::cert_find_by_email_stmt(&conn)?;
 
         cert_query!(stmt, [ &email ], StoreError::NoMatches(email.into()))
     }
@@ -647,7 +660,8 @@ impl<'a> Store<'a> for Pep {
         tracer!(TRACE, "Pep::fingerprints");
 
         let inner = || -> Result<Vec<Fingerprint>> {
-            let mut stmt = Self::cert_list_stmt(&self.conn)?;
+            let conn = self.conn.lock().unwrap();
+            let mut stmt = Self::cert_list_stmt(&conn)?;
 
             let rows = wrap_err!(
                 stmt.query_map([ ], |row: &Row| {
@@ -693,7 +707,8 @@ impl<'a> Store<'a> for Pep {
         tracer!(TRACE, "Pep::certs");
 
         let inner = || -> Result<Vec<_>> {
-            let mut stmt = Self::cert_all_stmt(&self.conn)?;
+            let conn = self.conn.lock().unwrap();
+            let mut stmt = Self::cert_all_stmt(&conn)?;
             cert_query!(stmt, [ ], StoreError::NoMatches("EOF".into()))
         };
 
@@ -708,8 +723,8 @@ impl<'a> Store<'a> for Pep {
 }
 
 impl<'a> StoreUpdate<'a> for Pep {
-    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
-                 merge_strategy: &mut dyn MergeCerts<'a>)
+    fn update_by(&self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &dyn MergeCerts<'a>)
         -> Result<Arc<LazyCert<'a>>>
     {
         tracer!(TRACE, "Pep::update_by");
@@ -717,8 +732,9 @@ impl<'a> StoreUpdate<'a> for Pep {
         let fpr = cert.fingerprint();
         t!("Updating {}", fpr);
 
+        let mut conn = self.conn.lock().unwrap();
         let tx = wrap_err!(
-            self.conn.transaction(),
+            conn.transaction(),
             UnknownDbError,
             "starting transaction"
         )?;
@@ -1311,7 +1327,7 @@ mod tests {
             std::fs::write(&filename, data)?;
         }
 
-        let mut pep = Pep::open(Some(filename)).expect("can open");
+        let pep = Pep::open(Some(filename)).expect("can open");
 
         // Certs.
         assert_eq!(records.len(), pep.certs().count());
